#include "imageviewer.h"
#include "imageloader.h"
#include <qpainter.h>
#include <qapplication.h>
#include <qpixmap.h>
#include <qdir.h>
#include <qfileinfo.h>
#include <qimage.h>
#include <qtimer.h>
#include <qcursor.h>
#include <stdlib.h>
#include <stdio.h>
#include <ksimpleconfig.h>
#include <klocale.h>
#include <kglobal.h>
#include <kiconloader.h>
#include <kmessagebox.h>
#include <kstandarddirs.h>

ImageView::ImageView(QWidget *parent, const char *name)
    : QScrollView(parent, name, WRepaintNoErase| WResizeNoErase |
                  WStaticContents)
{
    qWarning("Creating new image viewer!");
    initialized = true;
    hasMimetype = false;
    hasSizeCaption = false;
    hasError = false;
    pix = NULL;
    image = NULL;
    job = NULL;
    loader = NULL;
    preview = NULL;
    buffer = NULL;
    start_y = 10;
    textWidth = 0;
    finished = false;
    hasHandCursor = false;
    currentHighlight = -1;

    setFocusPolicy(ClickFocus);
    viewport()->setBackgroundColor(Qt::white);
    viewport()->setMouseTracking(true);
    loadConfig();
}

ImageView::~ImageView()
{
    qWarning("In ImageView destructor");
    delete fm;
    if(job){
        job->kill();
        job = NULL;
    }

    if(buffer){
        if(!buffer->remove())
            qWarning("Unable to delete buffer file %s!",
                     (const char *)QFile::encodeName(buffer->name()));
        delete buffer;
    }
    if(pix)
        delete pix;
    if(image)
        delete image;
    if(loader)
        delete loader;
    if(preview)
        delete preview;
    qWarning("ImageView destructor finished");
}

void ImageView::loadConfig()
{
    KConfig *config = new KSimpleConfig("pixierc");
    config->setGroup("Thumbnails");
    hotNameList = config->readListEntry("Names");
    hotPathList = config->readListEntry("Paths");
    hotNameList.prepend(i18n("Home"));
    hotPathList.prepend(QDir::homeDirPath());
    delete config;

    config = KGlobal::config();
    config->setGroup("General");
    QColor defaultColor(0, 0, 192);
    linkColor = config->readColorEntry("linkColor", &defaultColor);

    linkFont = font();
    linkFont.setUnderline(true);
    fm = new QFontMetrics(linkFont);
}

void ImageView::reloadConfig()
{
    if(!fileStr.isEmpty()){
        dupePathList.clear();
        dupeNameList.clear();
        dupeRectList.clear();
        totalLinkRect = QRect();
        start_y = 10;
        textWidth = 0;
        if(hasHandCursor)
            qApp->restoreOverrideCursor();
        hasHandCursor = false;
        loadConfig();
        checkForDupes(fileStr);
        repaintContents(false);
    }
}

void ImageView::save(const QString &fn)
{
    QFileInfo fi(fn);
    if(isLocalFile || !buffer){
        qWarning("Copying from original location.");
        KURL destURL(QString("file:" + fi.absFilePath()));
        KIO::Job *copyJob = KIO::copy(curURL, destURL);
        connect(copyJob, SIGNAL(result(KIO::Job *)),
                this, SLOT(slotCopyJobResult(KIO::Job *)));
        return;
    }
    else{
        qWarning("Using buffer");
        KURL srcURL("file:"+QFileInfo(*buffer).absFilePath());
        KURL destURL(QString("file:" + fi.absFilePath()));
        KIO::Job *copyJob = KIO::copy(srcURL, destURL);
        connect(copyJob, SIGNAL(result(KIO::Job *)),
                this, SLOT(slotCopyJobResult(KIO::Job *)));
    }
}

void ImageView::checkForDupes(const QString &fileName)
{
    QStringList::Iterator it = hotPathList.begin();
    QStringList::Iterator nameIt = hotNameList.begin();
    for(; it != hotPathList.end(); ++it, ++nameIt){
        if(QFile::exists((*it)+ "/" + fileName)){
            dupePathList.append(*it);
            dupeNameList.append(*nameIt);
            QString textStr = i18n("Duplicate filename in hotlist group ");
            textStr += (*nameIt);
            textStr += i18n(". Click to view.");

            int w = fm->width(textStr);
            int h = fm->lineSpacing();
            if(w > textWidth)
                textWidth = w;
            QRect r(10, start_y, w, h);
            dupeRectList.append(r);
            totalLinkRect = totalLinkRect.unite(r);
            start_y += h;
        }
    }
}

bool ImageView::closeURL()
{
    if(job){
        job->kill();
        job = NULL;
    }
    if(loader){
        delete loader;
        loader = NULL;
    }
    if(buffer)
        buffer->close();

    finished = true;
    return(true);
}

void ImageView::reset()
{
    if(job)
        job->kill();
    if(buffer){
        if(!buffer->remove())
            qWarning("Unable to delete buffer file %s!",
                     (const char *)QFile::encodeName(buffer->name()));
        delete buffer;
    }
    if(pix)
        delete pix;
    if(image)
        delete image;
    if(loader)
        delete loader;
    if(preview)
        delete preview;
    buffer = NULL;
    pix = NULL;
    image = NULL;
    loader = NULL;
    preview = NULL;
    job = NULL;
    start_y = 10;
    textWidth = 0;
    finished = false;
    currentHighlight = -1;
    hasMimetype = false;
    hasSizeCaption = false;
    hasError = false;
    mimeStr = QString::null;
    fileStr = QString::null;
    dupePathList.clear();
    dupeNameList.clear();
    dupeRectList.clear();
    totalLinkRect = QRect();
    if(hasHandCursor)
        qApp->restoreOverrideCursor();
    hasHandCursor = false;
}

bool ImageView::openURL(const KURL &url, const QString &mimetype, bool reload)
{
    if(initialized)
        reset();

    qWarning("In openURL for %s, %s", url.url().latin1(), mimetype.latin1());
    if(!mimetype.isEmpty()){
        mimeStr = mimetype;
        hasMimetype = true;
    }
    curURL = url;
    isLocalFile = url.isLocalFile();
    fileStr = url.fileName();
    emit title(i18n("Loading") + ' ' + fileStr + i18n("..."));

    checkForDupes(fileStr);

    if(isLocalFile && mimeStr != "image/x-bmp"){
        qWarning("Using local file: %s", QString(url.directory() +
                                                 "/" + url.fileName()).latin1());
        return(openFile(url.directory() + "/" +url.fileName()));
    }
    else{
        if(mimeStr == "image/jpeg" || mimeStr == "image/pjpeg" ||
           mimeStr == "image/x-pjpeg" || mimeStr == "image/x-jpeg"){
            /* PJPEG is found in some Yahoo groups - no idea where the
             * mimetype came from but it's a normal JPEG */
            if(!image)
                image = new QImage();
            loader = new JPEGImageLoader(*image, this);
        }
        else if(mimeStr == "image/gif" || mimeStr == "image/x-gif"){
            if(!image)
                image = new QImage();
            loader = new GIFImageLoader(*image, this);
        }
        else if(mimeStr == "image/png" || mimeStr == "image/x-png"){
            if(!image)
                image = new QImage();
            loader = new PNGImageLoader(*image, this);
        }
        else if(mimeStr == "image/bmp" || mimeStr == "image/x-bmp"){
            if(!image)
                image = new QImage();
            loader = new BMPImageLoader(*image, this);
        }
        else{
            qWarning("Unknown mimetype, trying QImageIO");
            hasError = true;
        }
    }

    // Use KURL to figure out local filename because it encodes them
    // differently than QFile::encodeName(). Important for Japanese/Unicode
    // filenames.
    KURL bufferURL(QString("file:/" + fileStr));
    buffer = new QFile(locateLocal("tmp", bufferURL.fileName()));
    if(!buffer->open(IO_WriteOnly | IO_Truncate)){
        qWarning("Unable to open buffer file!");
        delete buffer;
        buffer = 0;
    }
    else
        qWarning("Using buffer %s", QFileInfo(*buffer).absFilePath().latin1());

    job = KIO::get(url, reload, false);
    connect(job, SIGNAL(totalSize(KIO::Job *, KIO::filesize_t)),
            this, SLOT(slotTotalSize(KIO::Job *, KIO::filesize_t)));
    connect(job, SIGNAL(data(KIO::Job *, const QByteArray &)),
            this, SLOT(slotReadData(KIO::Job *, const QByteArray &)));
    connect(job, SIGNAL(result(KIO::Job *)),
            this, SLOT(slotResult(KIO::Job *)));
    connect(job, SIGNAL(mimetype(KIO::Job *, const QString &)),
            this, SLOT(slotMimetype(KIO::Job *, const QString &)));
    job->addMetaData("cache", KIO::getCacheControlString(KIO::CC_Verify));
    emit started(job);
    return(true);
}

bool ImageView::openFile(const QString &file)
{

    qWarning("Loading local file, using nonincremental QImage loading.");
    emit started(NULL);
    image = new QImage(file);
    if(!image->isNull()){
        pix = new QPixmap();
        pix->convertFromImage(*image);
        resizeContents(pix->width()+10 > textWidth ?
                       pix->width()+10 : textWidth, pix->height()+start_y);
        finished = true;
        QString titleStr(fileStr);
        titleStr += QString().sprintf("(%dx%d)", pix->width(), pix->height());
        emit title(titleStr);
        emit status(i18n("Image loaded."));
        emit completed();
        delete image;
        image = NULL;
        return(true);
    }
    delete image;
    image = NULL;
    emit status(i18n("Error loading image!"));
    emit completed();
    return(false);
}

void ImageView::slotReadData(KIO::Job *, const QByteArray &input)
{
    if(!pix)
        pix = new QPixmap();

    if(!input.size())
        return;
    if(buffer)
        buffer->writeBlock(input);
    if(!hasError){
        hasError = !loader->readData(input.data(), input.size());
        if(hasError)
            qWarning("Error incrementally loading image, trying QImageIO!");
        else if(!hasSizeCaption && !image->isNull()){
            hasSizeCaption = true;
            QString titleStr(i18n("Loading") + ' ' + fileStr);
            titleStr += QString().sprintf("(%dx%d)", image->width(), image->height());
            emit title(titleStr);
        }
    }
}

void ImageView::slotTotalSize(KIO::Job *, KIO::filesize_t /*data_size*/)
{
    /*
    if(totalSize){
        // whoops, we already got data before knowing total size - reallocate
        qWarning("Resizing data buffer");
        data = (char *)realloc(data, data_size);
    }
    else{
        qWarning("Setting initial buffer size to %d bytes", data_size);
        data = (char *)malloc(data_size);
    }
    totalSize = data_size; */

}

void ImageView::slotMimetype(KIO::Job *, const QString &/*type*/)
{
    // We should always get this in openURL.
}

void ImageView::slotResult(KIO::Job *)
{
    bool hasJobError = false;
    qWarning("In slotResult");
    if(job->error()){
        job->showErrorDialog();
        emit canceled(job->errorString());
        hasJobError = true;
    }
    if(loader){
        delete loader;
        loader = NULL;
    }
    if(buffer)
        buffer->close();
    job = 0;
    finished = true;

    if(hasError){
        // there was an error with incremental loading - try QImageIO
        if(image){
            QImage tmpImage(QFileInfo(*buffer).absFilePath());
            if(!tmpImage.isNull()){
                image->reset();
                *image = tmpImage;
                hasError = false;
            }
        }
        else{
            image = new QImage(QFileInfo(*buffer).absFilePath());
            hasError = image->isNull();
        }
    }

    if(!hasError && image){
        updateScanlines(0, image->height());
        QString titleStr(fileStr);
        if(!image->isNull())
            titleStr += QString().sprintf("(%dx%d)", image->width(), image->height());
        emit title(titleStr);
        emit status(i18n("Image loaded."));
    }
    delete image;
    image = NULL;
    if(!hasJobError && !hasError)
        emit completed();
}

void ImageView::slotCopyJobResult(KIO::Job *copyJob)
{
    if(copyJob->error())
        copyJob->showErrorDialog();
    emit copyFinished();
}

void ImageView::drawContents(QPainter *p, int x, int y, int w, int h)
{
    if(pix){
        QRect updateR(x, y, w, h);
        QRect pixR(10, start_y, pix->width(), pix->height());
        QRect paintR;

        if(pixR.intersects(updateR)){
            paintR = updateR.intersect(pixR);
            p->drawPixmap(paintR.x(), paintR.y(), *pix,
                          paintR.x()-10, paintR.y()-start_y,
                          paintR.width(), paintR.height());
        }

        if(!pixR.contains(updateR)){
            QRegion region(updateR);
            region -= pixR;
            QMemArray<QRect>rects = region.rects();
            for(unsigned int i=0; i < rects.size(); ++i){
                p->fillRect(rects[i], Qt::white);
            }
            if(!dupePathList.isEmpty()){
                QStringList::Iterator it = dupeNameList.begin();
                QValueList<QRect>::iterator rIt = dupeRectList.begin();

                p->setPen(linkColor);
                p->setFont(linkFont);
                QString linkStr;
                for(int i=0; it != dupeNameList.end(); ++it, ++rIt, ++i){
                    linkStr = i18n("Duplicate filename in hotlist group ");
                    linkStr += (*it);
                    linkStr += i18n(". Click to view.");
                    p->drawText((*rIt), Qt::AlignLeft | Qt::AlignHCenter |
                                Qt::SingleLine, linkStr);
                    if(i == currentHighlight){
                        p->setPen(QPen(linkColor, 1, Qt::DotLine));
                        p->drawRect(*rIt);
                        p->setPen(linkColor);
                    }
                }
            }
        }
    }
    else
        p->fillRect(x, y, w, h, Qt::white);
}

void ImageView::updateScanlines(int start_row, int end_row)
{
    if(image->isNull() || image->width() == 0 || image->height() == 0)
        return;
    if(!pix){
        pix = new QPixmap();
        pix->convertFromImage(*image);
        resizeContents(image->width()+10 > textWidth ?
                       image->width()+10 : textWidth, image->height()+start_y);
        return;
    }
    if(pix->width() != image->width() || pix->height() != image->height()){
        pix->convertFromImage(*image);
        resizeContents(pix->width()+10 > textWidth ?
                       pix->width()+10 : textWidth, pix->height()+start_y);
        return;
    }

    int h = end_row-start_row;
    ++h;
    if(end_row-start_row < image->height()-64){
        QImage tmpImage(image->copy(0, start_row, image->width(), h));
        QPixmap tmpPix;
        tmpPix.convertFromImage(*image);
        bitBlt(pix, 0, start_row, &tmpPix, 0, start_row, image->width(), h,
               Qt::CopyROP, false);
    }
    else
        pix->convertFromImage(*image);
    repaintContents(0, start_y+start_row, image->width()+10, h, false);
}

void ImageView::contentsMousePressEvent(QMouseEvent *ev)
{
    if(ev->button() == Qt::LeftButton){
        if(totalLinkRect.contains(ev->pos())){
            QValueList<QRect>::iterator it = dupeRectList.begin();
            for(int i=0; it != dupeRectList.end(); ++it, ++i){
                if((*it).contains(ev->pos())){
                    if(!preview){
                        preview = new LocalImagePreview();
                        preview->resize(500, 300);
                    }
                    preview->show();
                    preview->setFile(dupePathList[i]+"/"+fileStr);
                }
            }
        }
    }
    else if(ev->button() == Qt::RightButton)
        emit rightClick(viewport()->mapToGlobal(contentsToViewport(ev->pos())));

}

void ImageView::contentsMouseMoveEvent(QMouseEvent *ev)
{
    int tmp = currentHighlight;
    if(hasHandCursor){
        if(!totalLinkRect.contains(ev->pos())){
            qApp->restoreOverrideCursor();
            hasHandCursor = false;
            currentHighlight = -1;
        }
        else{
            QValueList<QRect>::iterator it = dupeRectList.begin();
            for(int i=0; it != dupeRectList.end(); ++it, ++i){
                if((*it).contains(ev->pos())){
                    if(i == currentHighlight)
                        return;
                    else{
                        currentHighlight = i;
                        if(tmp != -1)
                            repaintContents(dupeRectList[tmp], false);
                        repaintContents((*it), false);
                        return;
                    }
                }
            }
            // cursor wasn't in a link
            qApp->restoreOverrideCursor();
            hasHandCursor = false;
            currentHighlight = -1;
        }
        if(tmp != -1)
            repaintContents(dupeRectList[tmp], false);
    }
    else if(!hasHandCursor && totalLinkRect.contains(ev->pos())){
        QValueList<QRect>::iterator it = dupeRectList.begin();
        for(int i=0; it != dupeRectList.end(); ++it, ++i){
            if((*it).contains(ev->pos())){
                qApp->setOverrideCursor(QCursor(Qt::PointingHandCursor));
                hasHandCursor = true;
                currentHighlight = i;
                if(tmp != -1)
                    repaintContents(dupeRectList[tmp], false);
                repaintContents((*it), false);
                return;
            }
        }
    }
}

LocalImagePreview::LocalImagePreview()
    : QScrollView(NULL, "localimagepreview", WRepaintNoErase| WResizeNoErase |
                  WStaticContents | WType_TopLevel)
{
    viewport()->setBackgroundColor(Qt::white);
    pix = NULL;
    setCaption(i18n("Local image preview"));
}

LocalImagePreview::~LocalImagePreview()
{
    if(pix)
        delete pix;
}

void LocalImagePreview::setFile(const QString &fn)
{
    filename = fn;
    QImage img(fn);
    if(!img.isNull()){
        if(!pix)
            pix = new QPixmap;
        pix->convertFromImage(img);
        resizeContents(pix->width(), pix->height());
        repaintContents(false);
    }
    else{
        if(pix)
            delete pix;
        pix = NULL;
        QFontMetrics fm(fontMetrics());
        QString str(i18n("Unable to load ") + fn);
        resizeContents(fm.width(str)+20, fm.lineSpacing()+20);
        repaintContents(false);
    }
    QFileInfo fi(fn);
    setCaption(fi.absFilePath());
}

void LocalImagePreview::drawContents(QPainter *p, int x, int y, int w, int h)
{
    QRect updateR(x, y, w, h);
    if(pix){
        updateR = updateR.intersect(QRect(0, 0, pix->width(), pix->height()));
        p->drawPixmap(updateR.x(), updateR.y(), *pix, updateR.x(),
                      updateR.y(), updateR.width(), updateR.height());
    }
    else{
        p->fillRect(x, y, w, h, Qt::white);
        if(!filename.isEmpty()){
            QFontMetrics fm(fontMetrics());
            QString str(i18n("Unable to load ") + filename);
            if(updateR.intersects(QRect(0, 0, fm.width(str)+20, fm.lineSpacing()+20))){
                p->drawText(0, 0, fm.width(str)+20, fm.lineSpacing()+20,
                            Qt::AlignCenter | Qt::SingleLine, str);
            }
        }
    }
}

