#include "imageloader.h"
#include <qimage.h>
#include <png.h>


PNGImageLoader *currentPNGLoader;
QImage *currentPNGImage;
int png_start_row;
int png_end_row;

extern "C"{
    png_structp png_ptr;
    png_infop info_ptr;
    bool hasPNGHeader;
    bool hasPNGError;

    void info_callback(png_structp png, png_infop info){
        if(hasPNGError || !currentPNGImage)
            return;
        png_uint_32 width;
        png_uint_32 height;
        int bit_depth;
        int color_type;
        png_get_IHDR(png, info, &width, &height, &bit_depth, &color_type,
                     0, 0, 0);

        if ( color_type == PNG_COLOR_TYPE_GRAY ) {
            // Black & White or 8-bit greyscale
            if ( bit_depth == 1 && info->channels == 1 ) {
                png_set_invert_mono( png );
                png_read_update_info( png, info );
                currentPNGImage->create( width, height, 1, 2, QImage::BigEndian );
                currentPNGImage->setColor( 1, qRgb(0,0,0) );
                currentPNGImage->setColor( 0, qRgb(255,255,255) );
                currentPNGImage->fill(0);
            } else {
                if ( bit_depth == 16 )
                    png_set_strip_16(png);
                else if ( bit_depth < 8 )
                    png_set_packing(png);
                int ncols = bit_depth < 8 ? 1 << bit_depth : 256;
                png_read_update_info(png, info);
                currentPNGImage->create(width, height, 8, ncols);
                for (int i=0; i<ncols; i++) {
                    int c = i*255/(ncols-1);
                    currentPNGImage->setColor( i, qRgba(c,c,c,0xff) );
                }
                currentPNGImage->fill(255);
                if ( png_get_valid(png, info, PNG_INFO_tRNS) ) {
                    int g = info->trans_values.gray;
                    if ( bit_depth > 8 ) {
                        // transparency support disabled for now
                    } else {
                        currentPNGImage->setAlphaBuffer( TRUE );
                        currentPNGImage->setColor(g, RGB_MASK & currentPNGImage->color(g));
                    }
                }
            }
        } else if ( color_type == PNG_COLOR_TYPE_PALETTE
                    && png_get_valid(png, info, PNG_INFO_PLTE)
                    && info->num_palette <= 256 )
        {
            // 1-bit and 8-bit color
            if ( bit_depth != 1 )
                png_set_packing( png );
            png_read_update_info( png, info );
            png_get_IHDR(png, info,
                         &width, &height, &bit_depth, &color_type, 0, 0, 0);
            currentPNGImage->create(width, height, bit_depth, info->num_palette,
                                    QImage::BigEndian);
            int i = 0;
            if ( png_get_valid(png, info, PNG_INFO_tRNS) ) {
                currentPNGImage->setAlphaBuffer( TRUE );
                while ( i < info->num_trans ) {
                    currentPNGImage->setColor(i, qRgba(
                                            info->palette[i].red,
                                            info->palette[i].green,
                                            info->palette[i].blue,
                                            info->trans[i]
                                           )
                                  );
                    i++;
                }
            }
            while ( i < info->num_palette ) {
                currentPNGImage->setColor(i, qRgba(
                                        info->palette[i].red,
                                        info->palette[i].green,
                                        info->palette[i].blue,
                                        0xff
                                       )
                              );
                i++;
            }
        } else {
            // 32-bit
            if ( bit_depth == 16 )
                png_set_strip_16(png);

            png_set_expand(png);

            if ( color_type == PNG_COLOR_TYPE_GRAY_ALPHA )
                png_set_gray_to_rgb(png);

            currentPNGImage->create(width, height, 32);
            currentPNGImage->fill(qRgb(255, 255, 255));

            // Only add filler if no alpha, or we can get 5 channel data.
            if (!(color_type & PNG_COLOR_MASK_ALPHA)
                && !png_get_valid(png, info, PNG_INFO_tRNS)) {
                png_set_filler(png, 0xff,
                               QImage::systemByteOrder() == QImage::BigEndian ?
                               PNG_FILLER_BEFORE : PNG_FILLER_AFTER);
                // We want 4 bytes, but it isn't an alpha channel
            } else {
                currentPNGImage->setAlphaBuffer(TRUE);
            }

            if ( QImage::systemByteOrder() == QImage::BigEndian ) {
                png_set_swap_alpha(png);
            }

            png_read_update_info(png, info);
        }

        // Qt==ARGB==Big(ARGB)==Little(BGRA)
        if ( QImage::systemByteOrder() == QImage::LittleEndian ) {
            png_set_bgr(png);
        }


        hasPNGHeader = true;
    }

    void row_callback(png_structp png, png_bytep new_row,
                      png_uint_32 row_num, int /*pass*/){
        //qWarning("In PNG row_callback row: %ud, image size: %d, %d",
        //         row_num, currentPNGImage->width(), currentPNGImage->height());
        if(hasPNGError || !currentPNGImage){
            qWarning("Returning early!");
            return;
        }
        unsigned char *old_row = currentPNGImage->scanLine(row_num);
        png_progressive_combine_row(png, old_row, new_row);
        if(png_start_row == -1){
            png_start_row = row_num;
            png_end_row = row_num;
        }
        else
            png_end_row = row_num;
    }

    void error_callback(png_structp, png_const_charp message){
        qWarning("libpng error: %s", message);
    }

    void end_callback(png_structp, png_infop){;}
}

PNGImageLoader::PNGImageLoader(QImage &image, ImageView *view)
    : ImageLoader(image, view)
{
    currentPNGLoader = this;
    currentPNGImage = &image;
    png_start_row=-1;
    png_end_row=-1;
    png_ptr = NULL;
    info_ptr = NULL;
    hasPNGHeader = false;
    hasPNGError = false;

    // initialize
    png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, 0, 0, 0);
    if(!png_ptr){
        hasPNGError = true;
        return;
    }
    png_set_error_fn(png_ptr, 0, 0, error_callback);
    png_set_compression_level(png_ptr, 9);

    info_ptr = png_create_info_struct(png_ptr);
    if(!info_ptr){
        hasPNGError = true;
        png_destroy_read_struct(&png_ptr, (png_infopp)NULL, (png_infopp)NULL);
        return;
    }
    if(setjmp((png_ptr)->jmpbuf)){
        hasPNGError = true;
        png_destroy_read_struct(&png_ptr, &info_ptr, 0);
        return;
    }

    // setup all our callbacks
    png_set_progressive_read_fn(png_ptr, (void *)this,
                                info_callback, row_callback, end_callback);
}

PNGImageLoader::~PNGImageLoader()
{
    qWarning("In PNGImageLoader destructor");
    if(png_ptr && info_ptr)
        png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);
    png_ptr = NULL;
    info_ptr = NULL;
    qWarning("PNGImageLoader destructor finished");
}

bool PNGImageLoader::readData(char *buffer, int size)
{
    if(hasPNGError){
        qWarning("Has Error!");
        return(false);
    }
    png_process_data(png_ptr, info_ptr, (png_byte *)buffer, size);
    if(png_start_row != -1)
        currentPNGLoader->imageUpdated(png_start_row, png_end_row);
    png_start_row = -1;
    return(true);
}

bool PNGImageLoader::error()
{
    return(hasPNGError);
}

